@@ -2,6 +2,8 @@ from qiskit import *
 import numpy as np
 from scipy.optimize import minimize
 
+from openquantumcomputing.Statistic import Statistic
+
 class QAOABase:
 
     def __init__(self):#,params = None):
@@ -13,10 +15,17 @@ class QAOABase:
         self.E=None
         self.Var=None
         self.current_depth=0 # depth at which local optimization has been done
-        self.angles_hist={}
-        self.costval={}
+        self.angles_hist={} # initial and final angles during optimization per depth
+        self.num_fval={} # number of function evaluations per depth
+        self.num_shots={} # number of total shots taken for local optimization per depth
+        self.costval={} # optimal cost values per depth
         self.gamma_grid=None
         self.beta_grid=None
+        self.stat=Statistic()
+
+        self.g_it=0
+        self.g_values={}
+        self.g_angles={}
 
 ################################
 # functions to be implemented:
@@ -45,25 +54,41 @@ class QAOABase:
 # generic functions
 ################################
 
-    def loss(self, angles, backend, depth, shots, noisemodel=None, params={}):
+    def loss(self, angles, backend, depth, shots, precision, noisemodel, params):
         """
         loss function
         :param params: additional parameters
         :return: an instance of the qiskti class QuantumCircuit
         """
-        circuit = []
-        circuit.append(self.createCircuit(angles, depth, params=params))
+        self.g_it+=1
+        circuit = self.createCircuit(angles, depth, params=params)
 
-        if backend.configuration().local:
-            job = execute(circuit, backend=backend, noise_model=noisemodel, shots=shots)
-        else:
-            job = start_or_retrieve_job(name+"_"+str(opt_iterations), backend, circuit, options={'shots' : shots})
+        n_target=shots
 
-        e,v = self.measurementStatistics(job, params=params)
+        self.stat.reset()
+
+        shots_taken=0
+
+        for i in range(3):
+            if backend.configuration().local:
+                job = execute(circuit, backend=backend, noise_model=noisemodel, shots=shots)
+            else:
+                job = start_or_retrieve_job(name+"_"+str(opt_iterations), backend, circuit, options={'shots' : shots})
+            shots_taken+=shots
+            _,_ = self.measurementStatistics(job, params=params)
+            v=self.stat.get_Variance()
+            shots=int((np.sqrt(v)/precision)**2)-shots_taken
+            if shots<=0:
+                break
+
+        self.num_shots['d'+str(self.current_depth+1)]+=shots_taken
+
+        self.g_values[str(self.g_it)] = -self.stat.get_E()
+        self.g_angles[str(self.g_it)] = angles.copy()
 
         #opt_values[str(opt_iterations )] = e[0]
         #opt_angles[str(opt_iterations )] = angles
-        return -e
+        return -self.stat.get_E()
 
     def measurementStatistics(self, job, params):
         """
@@ -78,37 +103,20 @@ class QAOABase:
             expectations = []
             variances = []
             for i, counts in enumerate(counts_list):
-                n_shots = jres.results[i].shots
-                E = 0
-                E2 = 0
+                self.stat.reset()
                 for string in counts:
-                    cost = self.cost(string, params)
-                    E += cost*counts[string]
-                    E2 += cost**2*counts[string];
-                if n_shots == 1:
-                    v = 0
-                else:
-                    E/=n_shots
-                    E2/=n_shots
-                    v = (E2-E**2)*n_shots/(n_shots-1)
-                expectations.append(E)
-                variances.append(v)
+                    # qiskit binary strings use little endian encoding, but our cost function expects big endian encoding. Therefore, we reverse the order
+                    cost = self.cost(string[::-1], params)
+                    self.stat.add_sample(cost, counts[string])
+                expectations.append(self.stat.get_E())
+                variances.append(self.stat.get_Variance())
             return expectations, variances
         else:
-            n_shots = jres.results[0].shots
-            E = 0
-            E2 = 0
             for string in counts_list:
-                cost = self.cost(string, params)
-                E += cost*counts_list[string]
-                E2 += cost**2*counts_list[string];
-            if n_shots == 1:
-                v = 0
-            else:
-                E/=n_shots
-                E2/=n_shots
-                v = (E2-E**2)*n_shots/(n_shots-1)
-            return E,v
+                # qiskit binary strings use little endian encoding, but our cost function expects big endian encoding. Therefore, we reverse the order
+                cost = self.cost(string[::-1], params)
+                self.stat.add_sample(cost, counts_list[string])
+            return self.stat.get_E(), self.stat.get_Variance()
 
     def hist(self, angles, backend, shots, noisemodel=None, params={}):
         depth=int(len(angles)/2)
@@ -119,6 +127,20 @@ class QAOABase:
             job = start_or_retrieve_job("hist", backend, circ, options={'shots' : shots})
         return job.result().get_counts()
 
+    def random_init(self, gamma_bounds,beta_bounds,depth):
+        """
+        Enforces the bounds of gamma and beta based on the graph type.
+        :param gamma_bounds: Parameter bound tuple (min,max) for gamma
+        :param beta_bounds: Parameter bound tuple (min,max) for beta
+        :return: np.array on the form (gamma_1, beta_1, gamma_2, ...., gamma_d, beta_d)
+        """
+        gamma_list = np.random.uniform(gamma_bounds[0],gamma_bounds[1], size=depth)
+        beta_list = np.random.uniform(beta_bounds[0],beta_bounds[1], size=depth)
+        initial = np.empty((gamma_list.size + beta_list.size,), dtype=gamma_list.dtype)
+        initial[0::2] = gamma_list
+        initial[1::2] = beta_list
+        return initial
+
 
     def interp(self, angles):
         """
@@ -131,11 +153,11 @@ class QAOABase:
         """
         depth=len(angles)
         tmp=np.zeros(len(angles)+2)
-        tmp[1:-1]=angles
+        tmp[1:-1]=angles.copy()
         w=np.arange(0,depth+1)
         return w/depth*tmp[:-1] + (depth-w)/depth*tmp[1:]
 
-    def sample_cost_landscape(self, backend, shots, noisemodel=None, params={}, verbose=True, angles={"gamma": [0,2*np.pi,20], "beta": [0,2*np.pi,20]}):
+    def sample_cost_landscape(self, backend, shots=1024, noisemodel=None, params={}, verbose=True, angles={"gamma": [0,2*np.pi,20], "beta": [0,2*np.pi,20]}):
         if verbose:
             print("Calculating Energy landscape for depth p=1...")
 
@@ -179,7 +201,7 @@ class QAOABase:
     def get_current_deptgh(self):
         return self.current_depth
 
-    def local_opt(self, angles0, backend, shots, noisemodel=None, params={}, method='Nelder-Mead'):
+    def local_opt(self, angles0, backend, shots, precision, noisemodel=None, params={}, method='COBYLA'):
         """
 
         :param angles0: initial guess
@@ -187,12 +209,20 @@ class QAOABase:
 
         depth=int(len(angles0)/2)
 
+        self.num_shots['d'+str(self.current_depth+1)]=0
         res = minimize(self.loss, x0 = angles0, method = method,
-                       args=(backend, depth, shots, noisemodel, params),
-                       options={'xatol': 1e-2, 'fatol': 1e-1, 'disp': True})
+                       args=(backend, depth, shots, precision, noisemodel, params))
         return res
 
-    def increase_depth(self, backend, shots, noisemodel=None, params={}, method='Nelder-Mead'):
+    def increase_depth(self, backend, shots=1024, precision=None, noisemodel=None, params={}, method='COBYLA'):
+        """
+        sample cost landscape
+
+        :param backend: backend
+        :param shots: if precision=None, the number of samples taken
+                      if precision!=None, the minimum number of samples taken
+        :param precision: precision to reach for expectation value based on error=variance/sqrt(shots)
+        """
 
         if self.current_depth == 0:
             if self.E is None:
@@ -200,14 +230,6 @@ class QAOABase:
             ind_Emin = np.unravel_index(np.argmin(self.E, axis=None), self.E.shape)
             angles0=np.array((self.gamma_grid[ind_Emin[1]], self.beta_grid[ind_Emin[0]]))
             self.angles_hist['d1_initial']=angles0
-
-            res = self.local_opt(angles0, backend, shots, noisemodel=noisemodel, params=params, method=method)
-            if not res.success:
-                raise Warning("Local optimization was not successful.", res)
-
-            self.angles_hist['d1_final']=res.x
-            self.costval['d1_final']=res.fun
-            self.current_depth=1
         else:
             gamma=self.angles_hist['d'+str(self.current_depth)+'_final'][::2]
             beta=self.angles_hist['d'+str(self.current_depth)+'_final'][1::2]
@@ -218,15 +240,20 @@ class QAOABase:
             angles0[1::2]=beta_interp
             self.angles_hist['d'+str(self.current_depth+1)+'_initial']=angles0
 
-            res = self.local_opt(angles0, backend, shots, noisemodel=noisemodel, params=params, method=method)
-            if not res.success:
-                raise Warning("Local optimization was not successful.", res)
-
-            self.angles_hist['d'+str(self.current_depth+1)+'_final']=res.x
-            self.costval['d'+str(self.current_depth+1)+'_final']=res.fun
-            self.current_depth+=1
-
+        self.g_it=0
+        self.g_values={}
+        self.g_angles={}
 
+        res = self.local_opt(angles0, backend, shots, precision, noisemodel=noisemodel, params=params, method=method)
+        if not res.success:
+            raise Warning("Local optimization was not successful.", res)
+        self.num_fval['d'+str(self.current_depth+1)]=res.nfev
+        print("cost(depth=",self.current_depth+1,")=", res.fun)
+
+        ind = min(self.g_values, key=self.g_values.get)
+        self.angles_hist['d'+str(self.current_depth+1)+'_final']=self.g_angles[ind]
+        self.costval['d'+str(self.current_depth+1)+'_final']=self.g_values[ind]
 
 
+        self.current_depth+=1
 
